/*
 * This file is part of SynTrans, which is an AI for OpenTTD
 *
 * SynTrans is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * SynTrans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SynTrans; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
// Add all the files for libraries here
import("util.superlib","SuperLib",19);
import("pathfinder.road", "RoadPathFinder", 3);
// Add all your files here
require("network.nut");
require("town.nut");
require("pathfinder.nut");
require("convoy.nut");

Result <- SuperLib.Result;
Log <- SuperLib.Log;
Helper <- SuperLib.Helper;
ScoreList <- SuperLib.ScoreList;
Money <- SuperLib.Money;
Tile <- SuperLib.Tile;
Direction <- SuperLib.Direction;
Engine <- SuperLib.Engine;
Vehicle <- SuperLib.Vehicle;
Station <- SuperLib.Station;
Airport <- SuperLib.Airport;
Industry <- SuperLib.Industry;
//Town <- SuperLib.Town;
Order <- SuperLib.Order;
OrderList <- SuperLib.OrderList;
Road <- SuperLib.Road;
RoadBuilder <- SuperLib.RoadBuilder;

aiInstance <- null;
 
 // Add all the files for the entire project here
 // ex. require("globals.nut");
 
 class SynTrans extends AIController {
	networks = null;
	airnet = null;
	enableAir = null;
	saved = null;
	center = null;
	constructor() {
		networks = [];
		enableAir = false;
		saved = false;
		//airnet = AirNet();
	}
	function Start();
	function Save();
	function Load(version,data);
}

function SynTrans::Start()
{
	local oldMoney = 0;
	AIController.Sleep(5);
	aiInstance = this;
	AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
	Log.Info("SynTrans Started",Log.LVL_SUB_DECISIONS);
	if(saved == true) {
		Log.Info("Loading networks...",Log.LVL_SUB_DECISIONS);
		local stationlist = null;
		// Load RoadNetwork
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_ROAD)) {
			local depots = AIDepotList(AITile.TRANSPORT_ROAD);
			local newNet = Network(AIVehicle.VT_ROAD);
			stationlist = AIStationList(AIStation.STATION_BUS_STOP);
			foreach(station, _ in stationlist)
			{
				local town = AIStation.GetNearestTown(station);
				Log.Info("Processing town " + AITown.GetName(town));
				//newNet.AddTown(town);
				local theTown = Town(town,AIStation.STATION_BUS_STOP);
				theTown.stations.append(station);
				// Find the depot for the station... it should not be that far away
				//local tiles = Tile.MakeTileRectAroundTile(AIStation.GetLocation(station),20);
				foreach(depot, _ in depots) {
					local dtilex = AIMap.GetTileX(depot);
					Log.Info("Depot Tile X: " + dtilex);
					local dtiley = AIMap.GetTileY(depot);
					Log.Info("Depot Tile Y: " + dtiley);
					local stilex = AIMap.GetTileX(AIStation.GetLocation(station));
					Log.Info("Station Tile X: " + stilex);
					local stiley = AIMap.GetTileY(AIStation.GetLocation(station));
					Log.Info("Station Tile Y: " + stiley);
					for(local i = 10;i < 1000;i += 5) {
						if(stilex - i < dtilex && stilex + i > dtilex && stiley - i < dtilex && stiley + i > dtiley) {
							theTown.depot = depot;
							continue;
						}
					}
					if(!theTown.depot) Log.Error("Could not find a depot for this city.  I wonder why...");
				}
				newNet.towns.append(theTown);
			}
			newNet.defaultEngine = newNet.ChooseVehicle();
			networks.append(newNet);
		}
		// Load AirNetwork
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_AIR)) {
			airnet = Network(AIVehicle.VT_AIR);
			stationlist = AIStationList(AIStation.STATION_AIRPORT);
			foreach(station, _ in stationlist)
			{
				local town = AIStation.GetNearestTown(station);
				Log.Info("Processing town " + AITown.GetName(town));
				//airnet.AddTown(town);
				local theTown = Town(town,AIStation.STATION_AIRPORT);
				theTown.stations.append(station);
				theTown.depot = Airport.GetHangarTile(station);
				airnet.towns.append(theTown);
			}
			airnet.defaultEngine = airnet.ChooseVehicle();
		}
	}
	else {
		if(Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_ROAD) && Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_AIR)) {
			Log.Error("This is a road and air only AI.  Please enable AI these building in Advanced settings!");
		}
		// Name the company
		if(!AICompany.SetName("SynapticTransport Inc.")) {
			if(!AICompany.SetName("Synchronized Transport Inc.")) {
				if(!AICompany.SetName("Another Synaptic Transport Inc.")) {
					local i = 1;
					while(!AICompany.SetName("Synaptic Impersonator #" + i)) {
						i = i + 1;
					}
				}
			}
		}
		local built = false;
		local tries = 0;
		Log.Info("Building HQ...",Log.LVL_SUB_DECISIONS);
		// Build the HQ
		local townlist = AITownList();
		townlist.Valuate(AITown.GetPopulation);
		townlist.Sort(AIList.SORT_BY_VALUE, false);
		foreach(hqtown, _ in townlist) {
			Log.Info("Attempting to build HQ in " + AITown.GetName(hqtown),Log.LVL_SUB_DECISIONS);
			local skip = false;
			local tilelist = Tile.GetTownTiles(hqtown);
			tilelist.Valuate(AITile.GetClosestTown);
			tilelist.KeepValue(hqtown);
			// Scan for any company HQs and go in a different city if one is there
			for(local i = 0;i < 16;i++) {
				if(AICompany.ResolveCompanyID(i) == AICompany.COMPANY_INVALID) continue;
				local tile = AICompany.GetCompanyHQ(i);
				if(tile != AIMap.TILE_INVALID) {
					if(AITile.GetClosestTown(tile) == hqtown) {
						Log.Info("Found another player's HQ in town " + AITown.GetName(hqtown));
						skip = true;
						break;
					}
				}
			}
			if(skip) continue;
			tilelist.Valuate(AITile.IsBuildableRectangle,2,2);
			tilelist.KeepValue(1);
			tilelist.Valuate(AIMap.DistanceManhattan,AITown.GetLocation(hqtown));
			tilelist.Sort(AIList.SORT_BY_VALUE, true);
			if(!AICompany.BuildCompanyHQ(tilelist.Begin())) {
				Log.Warning("First attempt to build HQ failed.");
			}
			else {
				built = true;
				center = hqtown;
			}
			while(!built && tries <= 5) {
				if(!AICompany.BuildCompanyHQ(tilelist.Next())) {
					Log.Warning("Could not build HQ!");
					tries++;
				}
				else {
					built = true;
					center = hqtown;
					break;
				}
			}
			if(built) break;
		}
		if(!center) {
			Log.Warning("Completely failed to build HQ :(");
		}
		// Make initial network
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_ROAD)) {
			local newNet = Network(AIVehicle.VT_ROAD, center);
			networks.append(newNet);
		}
		// Make (but dont start) Airnet
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_AIR)) airnet = Network(AIVehicle.VT_AIR);
	}
	local townlist;
	local networkStart;
	local event;
	while(true)
	{
		// Manage events...
		while(AIEventController.IsEventWaiting()) {
			event = AIEventController.GetNextEvent();
			switch(event.GetEventType()) {
				case AIEvent.AI_ET_ENGINE_PREVIEW:
					local preview = AIEventEnginePreview.Convert(event);
					preview.AcceptPreview();
					Log.Info("Accepted preview for vehicle " + preview.GetName());
					break;
				case AIEvent.AI_ET_VEHICLE_CRASHED:
					Log.Info("Caught vehicle crash report!", Log.LVL_SUB_DECISIONS);
					local crash = AIEventVehicleCrashed.Convert(event);
					// If it was a rail crossing and another player was not considerate enough to fix it or just plain want
					// to destroy our vehicles, we will fix the problem for them :)
					if(crash.GetCrashReason() == AIEventVehicleCrashed.CRASH_RV_LEVEL_CROSSING) {
						Log.Info("Trying to fix road vehicle crash with train...", Log.LVL_SUB_DECISIONS);
						local site = crash.GetCrashSite();
						local adjTiles = Tile.GetNeighbours4MainDir(site);
						local adjTiles2 = adjTiles;
						adjTiles2.Valuate(AIBridge.IsBridgeTile);
						adjTiles2.RemoveValue(0);
						foreach(bridge, _ in adjTiles2) {
							AIBridge.RemoveBridge(bridge);
							AIRail.SetCurrentRailType(AIRailTypeList().Begin());
							raildir = Tile.TurnDirClockwise90Deg(Tile.GetDirectionToAdjacentTile(site, bridge), 1);
							local railResult = null;
							if(raildir == Tile.DIR_NE || raildir == Tile.DIR_SW) railResult = AIRail.BuildRailTrack(bridge, AIRail.RAILTRACK_NE_SW);
							if(raildir == Tile.DIR_NW || raildir == Tile.DIR_SE) railResult = AIRail.BuildRailTrack(bridge, AIRail.RAILTRACK_NW_SE);
							if(!railResult) Log.Warning("Failed to build rail track: " + AIError.GetLastErrorString());
						}
						adjTiles.Valuate(AITile.HasTransportType, AITile.TRANSPORT_ROAD);
						adjTiles.RemoveValue(0);
						Log.Info("Beginning Conversion...");
						local result = Road.ConvertRailCrossingToBridge(site, adjTiles.Begin());
						local i = 0;
						while(!result.succeeded && i < 8) {
							if(result.permanently) {
								Log.Warning("Could not convert rail crossing to bridge!");
								break;
							}
							else {
								Log.Info("Trying again...", Log.LVL_SUB_DECISIONS);
								result = Road.ConvertRailCrossingToBridge(site, adjTiles.Begin());
								AIController.Sleep(250);
							}
							i++;
						}
					}
			}
		}
		/*// Give money to the networks
		local income = AICompany.GetBankBalance(AICompany.COMPANY_SELF) - oldMoney;
		Log.Info("My income is " + income, Log.LVL_SUB_DECISIONS);
		airnet.money += income * 0.7;
		income = income * 0.7;
		foreach(network in networks) {
			network.money += income / networks.len();
		}
		oldMoney = AICompany.GetBankBalance(AICompany.COMPANY_SELF);*/
		// finally, is the airnet ready to start?
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_AIR)) if(enableAir) airnet.ManageNetwork();
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_ROAD)) foreach(network in networks) network.ManageNetwork();
		// Should we build a new network?
		if(!Vehicle.IsVehicleTypeDisabledByAISettings(AIVehicle.VT_ROAD)) {
		if(AIController.GetSetting("max_networks") > networks.len()) {
				if(AICompany.GetBankBalance(AICompany.COMPANY_SELF) > 250000) {
					Log.Info("==========BUILDING NEW NETWORK!==========",Log.LVL_SUB_DECISIONS);
					townlist = AITownList();
					// take out the cities we already own
					foreach(network in networks) {
						foreach(town in network.towns) {
							townlist.RemoveItem(town.id);
						}
					}
					// now find the biggest city
					townlist.Valuate(AITown.GetPopulation)
					townlist.Sort(AIList.SORT_BY_VALUE, false);
					networkStart = townlist.Begin();
					local newNet = Network(AIVehicle.VT_ROAD,networkStart);
					networks.append(newNet);
				}
			}
		}
		if(AICompany.GetCompanyValue(AICompany.COMPANY_SELF) > 200000 && !enableAir) {
			Log.Info("==========STARTING AIRCRAFT SYSTEM==========", Log.LVL_SUB_DECISIONS);
			enableAir = true;
		}
		Log.Info("Waiting...",Log.LVL_SUB_DECISIONS);
		AIController.Sleep(100);
	}
}

function SynTrans::Save()
{
	local syntrans = {};
	return syntrans;
}

function SynTrans::Load(version,data)
{
	saved = true;
}
