/*
 * This file is part of SynTrans, which is an AI for OpenTTD
 *
 * SynTrans is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * SynTrans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SynTrans; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

class Town
{
	lastUsageCheck = null;
	lastNewVehicles = null;
	id = null;
	outerLayer = null;
	stations = null;
	depot = null;
	suspended = null;
	newStationTries = null;
	stationType = null;
	connections = null;
	constructor(townid,stype) {
		this.lastUsageCheck = 0;
		this.lastNewVehicles = 0;
		this.id = townid;
		this.outerLayer = true;
		this.stations = [];
		this.depot = null;
		this.newStationTries = 0;
		this.stationType = stype;
		this.connections = AIList();
	}
	function Begin(stationType);
	function BuildStation();
	function End();
}

function Town::Begin(stationType,useDTRS = false) {
	connections = AIList();
	Log.Info("Attempting to build station and depot");
	this.stationType = stationType;
	local stationLocation;
	local thestation;
	if(stationType == AIStation.STATION_BUS_STOP) {
		if(useDTRS) {
			thestation = Road.BuildMagicDTRSInTown(id,AIRoad.ROADTYPE_ROAD,4,Helper.GetPAXCargo(),Helper.GetPAXCargo());
			if(thestation.result) {
				Log.Warning("DEPOT: " + thestation.depot_tile);
				stationLocation = AIStation.GetLocation(thestation.station_id);
				depot = thestation.depot_tile;
			}
		}
		else {
			local cfunctions = ConvoyFunc();
			local theTile = cfunctions.FindLineBusStopLocation(id, Helper.GetPAXCargo(), false);
			if(!theTile) {
				Log.Warning("Could not find a good place to build in " + id);
				return false;
			}
			local result = cfunctions.BuildBusStop(theTile);
			if(result) stationLocation = theTile;
		}
	}
	else stationLocation = Airport.BuildAirportInTown(id,Airport.GetLastAvailableAirportType(Engine.GetEngine_PAXLink(0,AIVehicle.VT_AIR)),Helper.GetPAXCargo(),Helper.GetPAXCargo());
	if(!stationLocation) {
		Log.Warning("Could not build station in this city");
		return false;
	}
	local station;
	/*if(stationType != AIStation.STATION_BUS_STOP)*/ station = AIStation.GetStationID(stationLocation);
	//station = stationLocation;
	Log.Info("STATION NAME: " + AIStation.GetName(station));
	if(!AIStation.IsValidStation(station)) {
		Log.Info("Could not build the station");
		return false;
	}
	stations.append(station);
	if(stationType == AIStation.STATION_BUS_STOP && !useDTRS) depot = Road.BuildDepotNextToRoad(stationLocation, 50, 150);
	else if(stationType == AIStation.STATION_AIRPORT) depot = Airport.GetHangarTile(station);
	if(!depot) {
		Log.Warning("Could not build depot!");
		return false;
	}
	suspended = false;
	newStationTries = 0;
	return true;
}

function Town::BuildStation() {
	// Not yet implemented
	Log.Info("Empty Function");
}

function Town::End() {
	Log.Info("Leaving the town...");
	for(local i = 0;i < stations.len();i++) {
		//AIRoad.RemoveRoadStation(AIStation.GetLocation(stations[i]));
		if(stationType == AIStation.STATION_BUS_STOP) {
			while(AIStation.IsValidStation(stations[i])) AIRoad.RemoveRoadStation(AIStation.GetLocation(stations[i]));
		}
		else {
			AIAirport.RemoveAirport(AIStation.GetLocation(stations[i]));
		}
	}
	if(stationType == AIStation.STATION_BUS_STOP) AIRoad.RemoveRoadDepot(this.depot);
	this.id = null;
	Log.Info("Finished ending town...");
}
